/* ****************************************************************************
[SQATT] Socket Queue Architecture by TelTel

This is the header file of SQATT.
Copyright (C) 2004 TELTEL
                                                                                
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.
                                                                                
This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.
                                                                                
You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Please contact me, if you have question about SQATT.
                                                                                
e-mail: teltel_library@yahoo.co.jp
Web(Japanese): http://www.geocities.jp/sqatt_dev/
Web(English):  http://www.geocities.com/sqatt_develop/
**************************************************************************** */

#ifndef  SQATT_HEADER

#include  <ttLog.h>

#define  SQ_VERSION   1

#define  SQATT_BSIZE  8192

#define  SQATT_OK                    0
#define  SQATT_ERR                  -1
#define  SQATT_BUFFER_FULL          -2
#define  SQATT_NO_DATA              -3
#define  SQATT_RECEIVING            -4
#define  SQATT_CS_CLOSED            -5
#define  SQATT_TOO_LONG_DATA_REQ    -6
#define  SQATT_TIMEOUT              -7
#define  SQATT_TOO_LONG_REPLY       -8
#define  SQATT_INVALID_SEQUENCE     -9
#define  SQATT_BAD_QUEUE           -10
#define  SQATT_FD_TBL_FULL         -11
#define  SQATT_RECEIVED            -12
#define  SQATT_TOO_SMALL_BUFFER    -13
#define  SQATT_USREV               -14
#define  SQATT_SKIP_SEQ_NO         -15
#define  SQATT_DUP_SEQ_NO          -16
#define  SQATT_INVALID_LENGTH      -17
#define  SQATT_NO_REPLY_SERVER     -18
#define  SQATT_UNEXPECTED_REPLY    -19
#define  SQATT_INVALID_PARAMETER   -20
#define  SQATT_NOT_RECV_YET        -21
#define  SQATT_DATA_STILL_SAVED    -22
#define  SQATT_ALREADY_RECEIVED    -23
#define  SQATT_CONFLICTED_SQ_TYPE  -24
#define  SQATT_INVALID_VERSION     -25

#define  SQATT_SELECT       1
#define  SQATT_ISSET_SFD    2
#define  SQATT_ISSET_CFD    3
#define  SQATT_CS_OPEN      5
#define  SQATT_RECV_HEAD    6
#define  SQATT_RECV_BODY    7
#define  SQATT_SEND_REPLY   8
#define  SQATT_CS_CLOSE     9

#define  RV_HEAD         1
#define  RV_BODY         2
#define  USER_GET        3
#define  USER_REPLY      4

#define  MAX_SQTBL       32
#define  SQATT_IPADR     "127.0.0.1"

#define  SQATT_1_WAY_QUEUE   1
#define  SQATT_2_WAY_QUEUE   2

#define  LINE50 "--------------------------------------------------"

typedef struct
{
     int     fd_type ;             /* 0:sqatt fd , 1:user fd                 */
     int     ev_flag ;             /* 1: event received for this socket      */
     int     cfd ;                 /* File descripter of client socket       */
     int     r_phase ;             /* Data receiving phase                   */
     int     sv_size ;             /* Size of saved data in main buffer area */
     int     bd_size ;             /* Size of body part                      */
     int     ex_size ;             /* Except size of receive data            */
     int     seq_no ;              /* Sequential counter.                    */
     char    mbuf[ SQATT_BSIZE ] ; /* Main buffer area                       */
     char    tbuf[ SQATT_BSIZE ] ; /* Temporary buffer area                  */
}    SQTBL ;

typedef struct
{
     int     sfd       ;     /* file descripter of server process            */
     int     sq_type   ;     /* 1:1 way queue , 2:2way queue                 */
     int     p_phase   ;     /* Program control phase                        */
     int     cur_idx   ;     /* Sqtbl number of notified data receiving      */
     int     chk_idx   ;     /* Sqtbl number of data receiving check at next */
     int     sqtbl_cnt ;     /* Number of available sqtbl.                   */
     int     werrno    ;

     char    mrbuf[ SQATT_BSIZE ] ;   /* Main buffer area for reply          */
     char    trbuf[ SQATT_BSIZE ] ;   /* Temporary buffer area for reply     */

     fd_set  rfds ;
     fd_set  wfds ;
     fd_set  efds ;
     struct  sockaddr_in  sadr ;

     struct  timeval      ctimer ;    /* timer for sqatt client              */
     struct  timeval      stimer ;    /* timer for sqatt server              */

     SQTBL   sqtbl[ MAX_SQTBL ] ;
}    SQATT ;

typedef struct
{
     int     len     ;         /* total of length of header and body */
     int     version ;         /* version number of SQATT            */
     int     sq_type ;
     int     seq_no  ;
     int     rslt    ;
     char    msg_cd [  4 ] ;   /* all zero       */
     char    lhdt   [ 14 ] ;   /* yyyymmddhhnnss */
     char    reserve[  2 ] ;
}    SQATT_HEAD ;

static  LHAND  lhand ;

#define  SQATT_HEADER

#endif
