/* ****************************************************************************
[SQATT] Socket Queue Architecture by TelTel

This is the sub-program of SQATT.
Copyright (C) 2004 TELTEL

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.
                                                                                
This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.
                                                                                
You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Please contact me, if you have question about SQATT.
                                                                                
e-mail: sqatt_dev@yahoo.co.jp
Web(Japanese): http://www.geocities.jp/sqatt_dev/
Web(English):  http://uk.geocities.com/sqatt_develop/
**************************************************************************** */

#include   <stdio.h>
#include   <string.h>
#include   <stdlib.h>
#include   <stdarg.h>
#include   <fcntl.h>
#include   <sys/types.h>
#include   <sys/stat.h>
#include   <errno.h>

#include   <ttLog.h>

#include   <ttCommon.h>
#include   <ttDate_func.h>
#include   <ttFile_func.h>

#define  HEAD01  "    0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F"
#define  HEAD02  "  0123456789ABCDEF"
#define  LINE01  "-- -----------------------------------------------"
#define  LINE02  "  ----------------"

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
typedef  struct
{
     char    whex[ 2 ] ;
     char    filler ;
}    ABYTE ;

typedef  struct
{
     char    lno[ 2 ] ;
     char    filler1 ;
     ABYTE   hexs [ 16 ] ;
     char    filler2 ;
     char    chars[ 16 ] ;
     char    eol ;
}    DUMP_LINE ;

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
void    ttInitLog ( LHAND* , char* ) ;
void    ttILog    ( LHAND* , char* , ... ) ;
void    ttELog    ( LHAND* , char* , ... ) ;
void    ttWLog    ( LHAND* , char* , ... ) ;
void    ttXLog    ( LHAND* , char* , ... ) ;
void    ttDILog   ( LHAND* , void* , int , void* ) ;
void    ttDWLog   ( LHAND* , void* , int , void* ) ;
void    ttDELog   ( LHAND* , void* , int , void* ) ;
void    ttLogMsg  ( LHAND* , int , char , void* ) ;
void    ttSetLogSW( LHAND* , int , int ) ;
int     ttGetLogSW( LHAND* , int ) ;

static  void   ttDLog_main ( LHAND* , char , void* , int , void* ) ;
static  void   ttDLog_title( LHAND* , int  , char  , int , void* ) ;
static  void   ttDLog_out  ( LHAND* , int  , char  , int , DUMP_LINE* ) ;

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
void    ttInitLog( lhand , wfnid )
LHAND   *lhand ;
char    *wfnid ;
{
     int     i ;
     int     wlen ;
     int     tlen ;

     lhand->fd = -1 ;

     for( i = 0 ; i < TT_LOG_CHANNEL ; ++i )
     {
          lhand->logsw[ i ] = 0 ; /* Not display log message. */
     }

     wlen  = strlen( wfnid ) ;
     tlen  = sizeof( lhand->fnid ) ;

     if( 0 < wlen && wlen <= tlen )
     {
          lhand->fnid[ 0 ] = '_' ;
          memcpy( &lhand->fnid[ 1 ] , wfnid , wlen ) ;
     }
     else if( tlen < wlen )
     {
          lhand->fnid[ 0 ] = '_' ;
          memcpy( &lhand->fnid[ 1 ] , wfnid , tlen ) ;
     }
}

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
void    ttILog( LHAND *lhand , char *fmt , ... )
{
     char     emsg[ 256 ] ;
     va_list  vlist ;
                                                                                
     va_start( vlist , fmt ) ;
     vsprintf( emsg , fmt , vlist ) ;
                                                                                
     ttLogMsg( lhand , 1 , 'I' , emsg ) ;
}

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
void    ttWLog( LHAND *lhand , char *fmt , ... )
{
     char     emsg[ 256 ] ;
     va_list  vlist ;
                                                                                
     va_start( vlist , fmt ) ;
     vsprintf( emsg , fmt , vlist ) ;
                                                                                
     ttLogMsg( lhand , 2 , 'W' , emsg ) ;
}

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
void    ttELog( LHAND *lhand , char *fmt , ... )
{
     char     emsg[ 256 ] ;
     va_list  vlist ;
                                                                                
     va_start( vlist , fmt ) ;
     vsprintf( emsg , fmt , vlist ) ;
                                                                                
     ttLogMsg( lhand , 3 , 'E' , emsg ) ;
}

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
void    ttXLog( LHAND *lhand , char *fmt , ... )
{
     char     emsg[ 256 ] ;
     va_list  vlist ;
                                                                                
     va_start( vlist , fmt ) ;
     vsprintf( emsg , fmt , vlist ) ;
                                                                                
     ttLogMsg( lhand , 4 , 'X' , emsg ) ;
}

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
void    ttDILog( lhand , ptr , size , msg )
LHAND  *lhand ;
void   *ptr   ;
int     size  ;
void   *msg   ;
{
     ttDLog_main( lhand , 'I' , ptr , size , msg ) ;
}

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
void    ttDWLog( lhand , ptr , size , msg )
LHAND  *lhand ;
void   *ptr   ;
int     size  ;
void   *msg   ;
{
     ttDLog_main( lhand , 'W' , ptr , size , msg ) ;
}

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
void    ttDELog( lhand , ptr , size , msg )
LHAND  *lhand ;
void   *ptr   ;
int     size  ;
void   *msg   ;
{
     ttDLog_main( lhand , 'E' , ptr , size , msg ) ;
}

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
void    ttDLog_main( lhand , level , ptr , size , msg )
LHAND  *lhand ;
char    level ;
void   *ptr   ;
int     size  ;
void   *msg   ;
{
     int     i ;
     int     idx ;
     int     lcnt ;
     int     log_sw ;
     char    wchar ;
     char    wdump[ 5 ] ;
     DUMP_LINE   dline ;

     memset( &dline , 0x20 , sizeof( DUMP_LINE )) ;
     idx  = 0 ;
     lcnt = 0 ;

     switch( level )
     {
          case 'E' : log_sw = 3 ; break ;
          case 'W' : log_sw = 2 ; break ;
          default  : log_sw = 1 ; break ;
     }

     for( i = 0 ; i < size ; ++i )
     {
          memset( wdump , 0x00 , sizeof( wdump )) ;
          wchar = *(( char* )( ptr + i )) ;
          sprintf( wdump , "%02X" , wchar ) ;
          memcpy( dline.hexs[ idx ].whex , wdump , 2 ) ;

          if(( 0x20 <= wchar && wchar <= 0x5F ) ||
             ( 0x61 <= wchar && wchar <= 0x7D ))
          {
               dline.chars[ idx ] = wchar ;
          }
          else
          {
               dline.chars[ idx ] = '.' ;
          }

          idx += 1 ;

          if( idx >= 16 )
          {
               if( lcnt % 16 == 0 )
               {
                    ttDLog_title( lhand , log_sw , level , size , msg ) ;
               }

               ttDLog_out( lhand , log_sw , level , lcnt , &dline ) ;
               lcnt += 1 ;
               idx   = 0 ;
          }
     }

     if( idx > 0 )
     {
          if( lcnt % 16 == 0 )
          {
               ttDLog_title( lhand , log_sw , level , size , msg ) ;
          }

          ttDLog_out( lhand , log_sw , level , lcnt , &dline ) ;
          lcnt += 1 ;
          idx   = 0 ;
     }
}

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
void   ttDLog_title( lhand , log_sw , level , size , msg )
LHAND *lhand  ;
int    log_sw ;
char   level  ;
int    size   ;
void  *msg    ;
{
     char    wmsg[ 256 ] ;

     memset  ( wmsg  ,  0x00  , sizeof( wmsg )) ;
     sprintf ( wmsg  , "Data Size : %5d Bytes  Comment: %s" , size , msg ) ;
     ttLogMsg( lhand , log_sw , level , wmsg ) ;

     memset  ( wmsg  ,  0x00  , sizeof( wmsg )) ;
     sprintf ( wmsg  , "%s%s" , HEAD01 , HEAD02 ) ;
     ttLogMsg( lhand , log_sw , level  , wmsg ) ;

     memset  ( wmsg  ,  0x00  , sizeof( wmsg )) ;
     sprintf ( wmsg  , "%s%s" , LINE01 , LINE02 ) ;
     ttLogMsg( lhand , log_sw , level  , wmsg ) ;
}

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
void    ttDLog_out( lhand , log_sw , level , lcnt , dline )
LHAND  *lhand  ;
int     log_sw ;
char    level  ;
int     lcnt   ;
DUMP_LINE  *dline ;
{
     char    wlcnt[ 5 ] ;

     memset ( wlcnt , 0x00 , sizeof( wlcnt )) ;
     sprintf( wlcnt , "%02X" , lcnt ) ;
     memcpy ( dline->lno , wlcnt , sizeof( dline->lno )) ;
     dline->eol = 0x00 ;

     ttLogMsg( lhand , log_sw , level , dline ) ;
     memset  ( dline , 0x20   , sizeof( DUMP_LINE )) ;
}

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
void    ttLogMsg( lhand , ch_no , level , msg )
LHAND   *lhand ;
int     ch_no  ;
char    level  ;
void    *msg   ;
{
     static  int     fd = 0 ;

     int     rslt ;
     int     werrno ;
     char    fname[ 256 ] ;
     char    buf  [ 256 ] ;
     char    wdate[  10 ] ;

     if( ttGetLogSW( lhand , ch_no ) == 0 )   return ;

     sprintf( buf , "%s %c %s\n" , ttTime( ':' ) , level , msg ) ;
     rslt = 0 ;

     memset( wdate , 0x00 , sizeof( wdate )) ;
     strcpy( wdate , ttDate2( '-' )) ;

     /* If log file has already opend and date change , */
     /* log file is closed.                             */

     if(( lhand->fd != -1 ) &&
        ( memcmp( wdate , lhand->log_date , sizeof( lhand->log_date )) != 0 ))
     {
          close( lhand->fd ) ;
          lhand->fd = -1 ;
     }

     if( lhand->fd == -1 )
     {
          memset( lhand->log_date , 0x00  , sizeof( lhand->log_date )) ;
          memcpy( lhand->log_date , wdate , sizeof( lhand->log_date )) ;

          memset ( fname , 0x00 , sizeof( fname )) ;
          sprintf( fname , "%s/log/log%s.%s" ,
                   ttHomeDir() , lhand->fnid , wdate ) ;

          if( ttFileExists( fname ) == 0 )
          {
               ttCreateFile( fname ) ;
          }

          lhand->fd = open( fname , O_RDWR ) ;
     }

     if( lhand->fd != -1 )
     {
          rslt   = lseek( lhand->fd , 0 , SEEK_END ) ;
          werrno = errno ;

          if( rslt == -1 )
          {
               printf( "Lseek error !! fd:%03d , Src:%s , Line:%05d\n" ,
                        lhand->fd , __FILE__ , __LINE__ ) ;
               printf( "errno:%4d , %s\n" , werrno , strerror( werrno )) ;
               printf( "Msg: %s\n" , msg ) ;
          }
          else
          {
               rslt = write( lhand->fd , buf , strlen( buf )) ;

               if( rslt <= 0 )
               {
                    perror( "Write Error !!" ) ;
                    printf( "Write Error !! Code:%4d , Src:%s , Line:%05d\n" ,
                             rslt , __FILE__ , __LINE__ ) ;
               }
          }
     }
     else
     {
          rslt = -1 ;
     }
}

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
void    ttSetLogSW( lhand , ch_no , value )
LHAND  *lhand ;
int     ch_no ;
int     value ;
{
     if(( 0 <= ch_no && ch_no < TT_LOG_CHANNEL ) &&
        ( value == 0 || value == 1 ))
     {
          lhand->logsw[ ch_no ] = value ;
     }
}

/* ************************************************************ */
/* *                                                          * */
/* ************************************************************ */
int     ttGetLogSW( lhand , ch_no )
LHAND  *lhand ;
int     ch_no ;
{
     return( lhand->logsw[ ch_no ]) ;
}
