/* ****************************************************************************
[SQATT] Socket Queue Architecture by TelTel

SQATT is the library program providing the IPC(Inter Process Communication)
function.

Copyright (C) 2004 TELTEL

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Please contact me, if you have question about SQATT.

e-mail: teltel_library@yahoo.co.jp
Web(Japanese): http://www.geocities.jp/sqatt_dev/
Web(English):  http://www.geocities.com/sqatt_develop/
**************************************************************************** */

#include   <stdio.h>
#include   <string.h>
#include   <stdlib.h>
#include   <stdarg.h>

#include   <sys/types.h>
#include   <sys/socket.h>
#include   <sys/time.h>
#include   <sys/ioctl.h>
#include   <netinet/in.h>
#include   <arpa/inet.h>
#include   <unistd.h>
#include   <signal.h>
#include   <errno.h>

#include   <ttCommon.h>
#include   <ttDate_func.h>
#include   <ttLog.h>
#include   <ttLog_func.h>

#include   <sqatt.h>

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */

int     sqatt_wait ( SQATT* ) ;
int     sqatt_sopen( SQATT* , int , int ) ;
int     sqatt_copen( SQATT* , int , int , int ) ;
int     sqatt_close( SQATT* ) ;
int     sqatt_recv ( SQATT* , int , int* , void* ) ;
int     sqatt_reply( SQATT* , int , void* ) ;
int     sqatt_send ( SQATT* , int , void* ) ;
int     sqatt_call ( SQATT* , int , void* , int , int* , void* ) ;

int     sqatt_add_usrfd( SQATT* , int ) ;
int     sqatt_del_usrfd( SQATT* , int ) ;
int     sqatt_get_usrfd( SQATT* ) ;

static  int     clear_sqtbl     ( SQTBL* ) ;
static  int     check_sqhead    ( SQATT* , SQATT_HEAD* ) ;
static  int     cs_recv         ( SQATT* , void* , int ) ;
static  int     csh_send        ( SQATT* , int , int , void* ) ;
static  int     get_blank_sqtbl ( SQATT* ) ;
static  int     get_size_in_sock( SQATT* ) ;
static  int     recv_reply      ( SQATT* , int , int* , void* ) ;
static  int     rp_recv         ( SQATT* , int , void* ) ;
static  int     set_fd_max      ( SQATT* ) ;
static  int     sqatt_accept    ( SQATT* ) ;
static  int     sqtbl_clrbuf    ( SQTBL* ) ;

static  void    edit_sqh( SQATT_HEAD* , SQATT* , int , int ) ;
static  void    countup_seqno( int* ) ;

static  void    dump_cfd( SQATT* ) ;
static  void    dump_sqh( SQATT_HEAD* ) ;
static  void    sqatt_signal( int ) ;

static  SQTBL   *cur_sqtbl( SQATT* ) ;
static  struct  timeval  *set_ctimeval( SQATT* ) ;
static  struct  timeval  *set_stimeval( SQATT* ) ;

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
int     sqatt_wait( sqatt )
SQATT   *sqatt     ;
{
     int     i ;
     int     end_flag ;
     int     rslt ;
     int     csh_size ;
     int     werrno ;
     int     rslt_sqatt_select ;
     int     widx ;
     int     cidx ;
     int     ridx ;
     int     wfd  ;
     int     closed_fd ;
     int     fd_max ;
     int     wk_body_size ;

     char    w_lhdt[   20 ] ;
     char    w_cd  [   20 ] ;
     char    buf   [ 4096 ] ;
     char    emsg  [  256 ] ;

     fd_set  tfds ;

     SQTBL    *sqtbl ;
     SQATT    xsqatt ;
     SQATT_HEAD   csh ;

     struct  timeval  *ptimer ;

     static  int     rslt_select = 0 ;

     end_flag          = 0 ;
     rslt_select       = 0 ;
     rslt_sqatt_select = 0 ;

     ttILog( &lhand , "%s" , LINE50 ) ;
     ttILog( &lhand , "[sqatt_wait()] Start. p_phase:%4d" , sqatt->p_phase ) ;
     dump_cfd( sqatt ) ;

     while( end_flag == 0 )
     {
          if( sqatt->p_phase == SQATT_SELECT )
          {
               ttILog( &lhand , "[SQATT_SELECT] Start." ) ;

               fd_max = set_fd_max  ( sqatt ) ;
               ptimer = set_stimeval( sqatt ) ;
               tfds   = sqatt->rfds ;

               ttXLog( &lhand , "timeval = sec:%5d. usec:%5d , fd_max = %5d" ,
                       sqatt->stimer.tv_sec , sqatt->stimer.tv_usec ,
                       fd_max ) ;
               ttILog( &lhand , "Waiting ..." ) ;

               rslt_select = select( fd_max , &tfds ,
                                     ( fd_set* )0 , ( fd_set* )0 , ptimer ) ;

               if( rslt_select == 0 )
               {
                    ttILog( &lhand , "Timeout occured at select()." ) ;
               }
               else if( 1 <= rslt_select )
               {
                    ttILog( &lhand , "Event was notified. rslt:%5d" ,
                                      rslt_select ) ;
                    sqatt->p_phase = SQATT_ISSET_SFD ;
               }
               else
               {
                    ttELog( &lhand , "select error. rslt:%5d , errno:%5d , %s" ,
                            rslt_select , werrno , strerror( werrno )) ;
                    rslt_sqatt_select = rslt_select ;
                    end_flag          = 1 ;
               }

               ttILog( &lhand , "[SQATT_SELECT] End." ) ;
          }
          else if( sqatt->p_phase == SQATT_ISSET_SFD )
          {
               ttILog( &lhand , "[SQATT_ISSET_SFD] Start." ) ;

               if( FD_ISSET( sqatt->sfd , &tfds ))
               {
                    rslt_select  -= 1 ;
                    sqatt->p_phase = SQATT_CS_OPEN ;
                    ttILog( &lhand ,
                            "Connection event received from client process." ) ;
               }
               else
               {
                    sqatt->p_phase = SQATT_ISSET_CFD ;
                    ttILog( &lhand , "Not find connection event." ) ;
               }

               ttILog( &lhand , "[SQATT_ISSET_SFD] End." ) ;
          }
          else if( sqatt->p_phase == SQATT_ISSET_CFD )
          {
               ttILog( &lhand , "[SQATT_ISSET_CFD] Start. chk_idx:%4d" ,
                                sqatt->chk_idx ) ;
               dump_cfd( sqatt ) ;
               cidx = sqatt->chk_idx ;
               ridx = -1 ;

               for( i = 0 ; i < MAX_SQTBL ; ++i )
               {
                    if( sqatt->sqtbl[ cidx ].cfd > 0 )
                    {
                         rslt = FD_ISSET( sqatt->sqtbl[ cidx ].cfd , &tfds ) ;

                         if( rslt != 0 )
                         {
                              ridx = cidx ;
                              i    = MAX_SQTBL ;
                         }
                    }

                    cidx = ( cidx >= MAX_SQTBL - 1 ) ? 0 : cidx + 1 ;
               }

               if( ridx == -1 ) 
               { 
                    /* Not notified at client socket. */

                    sqatt->p_phase = SQATT_SELECT ;
                    ttILog( &lhand , "Not notified at cfd." ) ;
               }
               else
               {
                    sqatt->sqtbl[ ridx ].ev_flag = 1 ;
                    rslt_select  -= 1 ;
                    sqatt->cur_idx = ridx ;

                    if( sqatt->sqtbl[ ridx ].fd_type == 0 )  /* sqatt fd */
                    {
                         if( sqatt->sqtbl[ ridx ].r_phase == RV_HEAD )
                         {
                              sqatt->p_phase = SQATT_RECV_HEAD ;
                              ttILog( &lhand ,
                                      "Notified event at cfd[%3d], fd:%3d" ,
                                      sqatt->cur_idx ,
                                      sqatt->sqtbl[ ridx ].cfd ) ;
                         }
                         else if( sqatt->sqtbl[ ridx ].r_phase == RV_BODY )
                         {
                              sqatt->p_phase = SQATT_RECV_BODY ;
                              ttILog( &lhand ,
                                      "Notified event at cfd[%3d], fd:%3d" ,
                                      sqatt->cur_idx ,
                                      sqatt->sqtbl[ ridx ].cfd ) ;
                         }
                         else
                         {
                              sqatt->p_phase = SQATT_CS_CLOSE ;
                              ttELog( &lhand ,
                                      "Data is still saved in buffer." ) ;
                              ttDELog( &lhand , sqtbl->mbuf , sqtbl->sv_size ,
                                       "Saved data" ) ;
                         }
                    }
                    else   /* user fd */
                    {
                         sqatt->p_phase    = SQATT_SELECT ;
                         rslt_sqatt_select = SQATT_USREV  ;
                         end_flag         = 1 ;

                         ttILog( &lhand ,
                                 "Notified user event at cfd[%3d], fd:%3d" ,
                                 sqatt->cur_idx , sqatt->sqtbl[ ridx ].cfd ) ;
                    }
               }

               sqatt->chk_idx = cidx ;
               ttILog( &lhand ,
                       "[SQATT_ISSET_CFD] End. Next chk_idx:%4d" ,
                       sqatt->chk_idx ) ;
          }
          else if( sqatt->p_phase == SQATT_CS_OPEN )
          {
               ttILog( &lhand , "[SQATT_CS_OPEN] Start." ) ;

               widx = get_blank_sqtbl( sqatt ) ;
               if( widx == -1 )
               {
                    /* Already connection full. Then connect as */
                    /* temporary and disconnect immediatery.    */

                    sqatt_accept( &xsqatt ) ;
                    xsqatt.cur_idx = 0 ;
                    xsqatt.sfd = sqatt->sfd ;

                    sqatt_close( &xsqatt ) ;
                    sqatt->p_phase = SQATT_ISSET_CFD ;

                    ttELog( &lhand , "SQATT connection full. Closed." ) ;
               }
               else
               {
                    sqatt->cur_idx   = widx ;
                    sqatt->sqtbl_cnt += 1 ;
                    sqatt->p_phase   = SQATT_ISSET_CFD ;

                    wfd = sqatt_accept( sqatt ) ;
                    FD_SET( wfd , &( sqatt->rfds )) ;

                    ttILog( &lhand ,
                            "Accept client. cfd[%3d], fd:%3d" ,
                            sqatt->cur_idx , wfd ) ;
               }

               ttILog( &lhand , "[SQATT_CS_OPEN] End." ) ;
          }
          else if( sqatt->p_phase == SQATT_RECV_HEAD )
          {
               ttILog( &lhand , "[SQATT_RECV_HEAD] Start." ) ;

               sqtbl    = cur_sqtbl( sqatt ) ;
               csh_size = sizeof( SQATT_HEAD ) ;
               memset( &csh , 0x00 , csh_size ) ;
               rslt = cs_recv( sqatt , &csh , csh_size ) ;

               if( rslt == SQATT_OK )
               {
                    ttILog( &lhand , "Completed receiving HEADER." ) ;
                    ttILog( &lhand , "total size: %5d bytes" , csh.len ) ;

                    rslt = check_sqhead( sqatt , &csh ) ;
                    if( rslt == SQATT_OK )
                    {
                         sqtbl->bd_size = csh.len - csh_size ;

                         if( sqtbl->bd_size > 0 )
                         {
                              sqatt->p_phase = SQATT_RECV_BODY ;
                              sqtbl->r_phase = RV_BODY ;
                              sqtbl_clrbuf( sqtbl ) ;

                              ttILog( &lhand , "Body part: %5d bytes" ,
                                       sqtbl->bd_size ) ;
                         }
                         else
                         {
                              sqatt->p_phase    = SQATT_SELECT   ;
                              sqtbl->r_phase    = USER_GET       ;
                              rslt_sqatt_select = SQATT_RECEIVED ;
                              end_flag          = 1 ;

                              ttILog( &lhand , "No body part." ) ;
                         }

                         dump_sqh( &csh ) ;  /* dump sqatt_head to log file */
                    }
                    else
                    {
                         sqatt->p_phase = SQATT_CS_CLOSE ;
                    }
               }
               else if( rslt == SQATT_RECEIVING )
               {
                    sqatt->p_phase = SQATT_SELECT ;
                    ttILog( &lhand ,
                            "Receiving size:%5d bytes" , sqtbl->sv_size ) ;
               }
               else if( rslt == SQATT_CS_CLOSED )
               {
                    sqatt->p_phase = SQATT_CS_CLOSE ;
                    ttILog( &lhand , "Client socket was closed by client." ) ;
               }
               else
               {
                    sqatt->p_phase = SQATT_CS_CLOSE ;
                    ttELog( &lhand , "Error occured. rslt:%5d" , rslt ) ;
               }

               ttILog( &lhand , "[SQATT_RECV_HEAD] End." ) ;
          }
          else if( sqatt->p_phase == SQATT_RECV_BODY )
          {
               ttILog( &lhand , "[SQATT_RECV_BODY] Start." ) ;

               sqtbl = cur_sqtbl( sqatt ) ;
               rslt  = cs_recv  ( sqatt , 0x00 , sqtbl->bd_size ) ;

               if( rslt == SQATT_OK )
               {
                    sqatt->p_phase    = SQATT_SELECT ;
                    sqtbl->r_phase    = USER_GET ;
                    rslt_sqatt_select = SQATT_RECEIVED ;
                    end_flag          = 1 ;

                    ttILog ( &lhand , "Completed receiving body part." ) ;
                    ttDILog( &lhand , sqtbl->mbuf , sqtbl->sv_size ,
                             "SQATT BODY" ) ;
               }
               else if( rslt == SQATT_RECEIVING  ||
                        rslt == SQATT_NO_DATA    )
               {
                    sqatt->p_phase = SQATT_SELECT ;
                    ttXLog( &lhand ,
                            "Receiving size:%5d bytes" , sqtbl->sv_size ) ;
               }
               else if( rslt == SQATT_CS_CLOSED )
               {
                    sqatt->p_phase = SQATT_CS_CLOSE ;
                    sqtbl->r_phase = RV_HEAD ;

                    ttELog( &lhand , "Client socket was closed by client." ) ;
               }
               else
               {
                    sqatt->p_phase = SQATT_CS_CLOSE ;
                    sqtbl->r_phase = RV_HEAD ;

                    ttELog( &lhand , "Error occured. rslt:%5d" , rslt ) ;
               }

               ttILog( &lhand , "[SQATT_RECV_BODY] End." ) ;
          }
          else if( sqatt->p_phase == SQATT_CS_CLOSE )
          {
               ttWLog( &lhand , "[SQATT_CS_CLOSE] Start." ) ;

               sqtbl = cur_sqtbl( sqatt ) ;
               if( sqtbl->sv_size > 0 )
               {
                    ttELog ( &lhand , "In the midst of receiving." ) ;
                    ttELog ( &lhand ,
                             "Expept size:   %5d bytes" , sqtbl->ex_size ) ;
                    ttELog ( &lhand ,
                             "received size: %5d bytes" , sqtbl->sv_size ) ;
                    ttDELog( &lhand ,
                             sqtbl->mbuf , sqtbl->sv_size , "Ignored data" ) ;
               }

               closed_fd = sqtbl->cfd ;
               FD_CLR( sqtbl->cfd , &( sqatt->rfds )) ;
               sqatt_close( sqatt ) ;
               sqatt->p_phase = SQATT_SELECT ;

               ttWLog( &lhand , "Close completed. fd:%2d" , closed_fd ) ;
               ttWLog( &lhand , "[SQATT_CS_CLOSE] End." ) ;
          }
          else
          {
               ttILog( &lhand , "Phase error. p_phase:%2d" , sqatt->p_phase ) ;
               exit( -1 ) ;
          }
     }

     dump_cfd( sqatt ) ;
     ttILog( &lhand , "[sqatt_wait()] End." ) ;

     return( rslt_sqatt_select ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
int     sqatt_sopen( sqatt , port_no , io_type )
SQATT   *sqatt  ;
int     port_no ;
int     io_type ;  /* 1:1 way queue , 2:2 way queue */
{
     int     i ;
     int     slen ;
     char    wfnid[ 32 ] ;

     memset ( wfnid , 0x00 , sizeof( wfnid )) ;
     sprintf( wfnid , "%05d_sv" , port_no ) ;

     ttInitLog ( &lhand , wfnid ) ;
     ttSetLogSW( &lhand , 1 , 1 ) ;  /* I */
     ttSetLogSW( &lhand , 2 , 1 ) ;  /* W */
     ttSetLogSW( &lhand , 3 , 1 ) ;  /* E */
     ttSetLogSW( &lhand , 4 , 1 ) ;  /* X */

     ttILog( &lhand , "[sqatt_sopen()] Start." ) ;

     sqatt->sfd = socket( AF_INET , SOCK_STREAM , 0 ) ;

     sqatt->sadr.sin_family      = AF_INET ;
     sqatt->sadr.sin_addr.s_addr = inet_addr( SQATT_IPADR ) ;
     sqatt->sadr.sin_port        = htons( port_no ) ;
     slen = sizeof( sqatt->sadr ) ;

     bind  ( sqatt->sfd , ( struct sockaddr * )&sqatt->sadr , slen ) ;
     listen( sqatt->sfd , 5 ) ;

     sqatt->sq_type   = io_type      ;  /* queue io type */
     sqatt->p_phase   = SQATT_SELECT ;
     sqatt->cur_idx   = -1 ;  /* Now, you have no data.       */
     sqatt->chk_idx   =  0 ;  /* Check the top of Sqtbl list. */
     sqatt->sqtbl_cnt =  0 ;  /* No available Sqtbl.          */

     for( i = 0 ; i < MAX_SQTBL ; ++i )
     {
          clear_sqtbl( &( sqatt->sqtbl[ i ])) ;
     }

     FD_ZERO( &( sqatt->rfds )) ;
     FD_ZERO( &( sqatt->wfds )) ;
     FD_ZERO( &( sqatt->efds )) ;
     FD_SET ( sqatt->sfd , &( sqatt->rfds )) ;

     sqatt->stimer.tv_sec  = 0 ;  /* timer for server (seconds)       */
     sqatt->stimer.tv_usec = 0 ;  /* timer for server (micro seconds) */

     ttILog( &lhand , "listen port:         %5d" , sqatt->sadr.sin_port ) ;
     ttILog( &lhand , "fd of server socket: %5d" , sqatt->sfd ) ;

     dump_cfd( sqatt ) ;
     ttILog( &lhand , "[sqatt_sopen()] End." ) ;

     return( 0 ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
int     sqatt_copen( sqatt , port_no , io_type , timer )
SQATT   *sqatt  ;
int     port_no ;
int     io_type ;   /* 1:1 way queue , 2:2 way queue */
int     timer   ;   /* timeout seconds               */
{
     int     rslt ;
     char    wfnid[ 32 ] ;
     SQTBL   *sqtbl ;
                                                                                
     memset ( sqatt , 0x00 , sizeof( SQATT )) ;
     memset ( wfnid , 0x00 , sizeof( wfnid )) ;
     sprintf( wfnid , "%05d_cl" , port_no ) ;
     ttInitLog ( &lhand , wfnid ) ;
     ttSetLogSW( &lhand , 1 , 1 ) ;  /* I */
     ttSetLogSW( &lhand , 2 , 1 ) ;  /* W */
     ttSetLogSW( &lhand , 3 , 1 ) ;  /* E */
     ttSetLogSW( &lhand , 4 , 1 ) ;  /* X */
     ttILog( &lhand , "[sqatt_copen()] Start." ) ;

     sqatt->sq_type = io_type ;
     sqatt->cur_idx = 0 ;
     sqtbl = cur_sqtbl( sqatt ) ;
     clear_sqtbl( sqtbl ) ;

     sqtbl->fd_type = 0 ;  /* sqatt fd */
     sqtbl->cfd     = socket( AF_INET , SOCK_STREAM , 0 ) ;
                                                                                
     sqatt->sadr.sin_family      = AF_INET ;
     sqatt->sadr.sin_addr.s_addr = inet_addr( SQATT_IPADR ) ;
     sqatt->sadr.sin_port        = htons( port_no ) ;
                                                                                
     rslt = connect( sqtbl->cfd, ( struct sockaddr * )&( sqatt->sadr ) ,
                     sizeof( sqatt->sadr )) ;

     sqatt->ctimer.tv_sec  = timer ;  /* timer for cliend (seconds)       */
     sqatt->ctimer.tv_usec = 0 ;      /* timer for client (micro seconds) */
                                                                                
     ttILog( &lhand , "listen port:         %5d" , sqatt->sadr.sin_port ) ;
     ttILog( &lhand , "fd of client socket: %5d" , sqtbl->cfd ) ;
     ttILog( &lhand , "reply timer:         %5d" , sqatt->ctimer.tv_sec ) ;

     ttILog( &lhand , "[sqatt_copen()] End." ) ;

     return( 0 ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  int     sqatt_accept( sqatt )
SQATT   *sqatt ;
{
     int     wsize ;
     int     rslt  ;
     struct  sockaddr_in  cadr ;
     SQTBL   *sqtbl ;

     ttILog( &lhand , "[sqatt_accept()] Start." ) ;

     sqtbl = &( sqatt->sqtbl[ sqatt->cur_idx ]) ;
     clear_sqtbl( sqtbl ) ;

     wsize = sizeof( struct sockaddr_in ) ;
     sqtbl->cfd = accept( sqatt->sfd , ( struct sockaddr *)&cadr , &wsize ) ;
     sqtbl->fd_type = 0 ;  /* 0:sqatt fd */
     sqtbl->r_phase = RV_HEAD ;

     ttILog( &lhand , "[sqatt_accept()] End." ) ;

     rslt = ( sqtbl->cfd == -1 ) ? -1 : sqtbl->cfd ;

     return( rslt ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
int     sqatt_close( sqatt )
SQATT   *sqatt ;
{
     SQTBL   *sqtbl ;

     sqtbl = cur_sqtbl( sqatt ) ;

     shutdown( sqtbl->cfd , SHUT_WR ) ;
     close   ( sqtbl->cfd ) ;

     clear_sqtbl( sqtbl ) ;

     return( 0 ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
int     sqatt_recv( sqatt , buf_size , data_size , data_buf )
SQATT   *sqatt ;
int     buf_size   ;
int     *data_size ;
void    *data_buf  ;
{
     int     rslt  ;
     int     rslt1 ;
     int     rslt2 ;
     SQTBL   *sqtbl ;

     ttILog( &lhand , "[sqatt_recv()] Start. buf_size:%5d" , buf_size ) ;

     sqtbl = cur_sqtbl( sqatt ) ;

     if( sqtbl->r_phase == USER_GET )
     {
          if( buf_size >= sqtbl->sv_size )
          {
               memset( data_buf , 0x00 , buf_size ) ;
               if( sqtbl->bd_size > 0 )
               {
                    memcpy( data_buf , sqtbl->mbuf , sqtbl->bd_size ) ;
               }
               *data_size = sqtbl->bd_size ;
               sqtbl_clrbuf( sqtbl ) ;
               sqtbl->r_phase = USER_REPLY ;
               rslt1 = SQATT_OK ;

               ttILog( &lhand ,
                       "The data was saved into user area. size:%8d bytes" ,
                       *data_size ) ;
          }
          else
          {
               ttELog( &lhand ,
                       "Too small buffer size. buffer:%5d , necessary:%5d" ,
                        buf_size , *data_size ) ;

               *data_size = sqtbl->sv_size ;
               sqtbl_clrbuf( sqtbl ) ;
               sqtbl->r_phase = RV_HEAD ;
               rslt1 = SQATT_TOO_SMALL_BUFFER ;
          }
     }
     else if( sqtbl->r_phase == RV_HEAD || sqtbl->r_phase == RV_BODY )
     {
          ttELog( &lhand , "Sqatt has not received data yet." ) ;
          rslt1 = SQATT_NOT_RECV_YET ;
     }
     else
     {
          ttELog( &lhand , "You have already received data." ) ;
          rslt1 = SQATT_ALREADY_RECEIVED ;
     }

     if( rslt1 == SQATT_OK )
     {
          if( sqatt->sq_type == 1 )  /* in the case of 1 way queue */
          {
               /* sleep( 10 ) ;  /* TEST */
               ttILog( &lhand , "One way queue, so send reply." ) ;
               rslt2 = csh_send( sqatt , rslt1 , 0 , 0x00 ) ;
               sqtbl->r_phase = RV_HEAD ;
          }
          else  /* In the case of 2 way queue */
          {
               rslt2 = SQATT_OK ;
               sqtbl->r_phase = USER_REPLY ;
               ttWLog( &lhand , "You must send reply message to client." ) ;
          }
     }
     else
     {
          rslt2 = SQATT_ERR ;
     }

     if( rslt1 == SQATT_OK && rslt2 == SQATT_OK )
     {
          rslt = SQATT_OK ;
     }
     else if( rslt1 == SQATT_OK && rslt2 != SQATT_OK )
     {
          rslt = rslt2 ;
     }
     else if( rslt1 != SQATT_OK && rslt2 == SQATT_OK )
     {
          rslt = rslt1 ;
     }
     else
     {
          rslt = rslt1 ;
     }

     ttILog( &lhand , "[sqatt_recv()] End. rslt:%4d" , rslt ) ;

     return( rslt ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
int     sqatt_reply( sqatt , body_len , sq_body )
SQATT   *sqatt   ;
int     body_len ;
void    *sq_body ;
{
     int     rslt ;
     SQTBL   *sqtbl ;

     ttILog ( &lhand ,
              "[sqatt_reply()] Start. body_len:%6d, sq_body: 0x%08X" ,
              body_len , sq_body ) ;

     sqtbl = cur_sqtbl( sqatt ) ;
     rslt  = csh_send ( sqatt , SQATT_OK , body_len , sq_body ) ;

     if( rslt == SQATT_OK )
     {
          ttILog( &lhand , "Completed replying. rslt:%5d" , rslt ) ;
     }
     else
     {
          ttELog( &lhand , "Aborted replying. rslt:%5d" , rslt ) ;
     }

     sqtbl->r_phase = RV_HEAD ;
     ttILog( &lhand , "[sqatt_reply()] End. rslt:%4d" , rslt ) ;

     return( rslt ) ;
}

/* ********************************************************************** */
/* *  cs_recv()                                                         * */
/* *                                                                    * */
/* *  Read data from socket until saved data equal to csd_size or       * */
/* *  socket is empty.                                                  * */
/* *                                                                    * */
/* ********************************************************************** */
static  int     cs_recv( sqatt , csd , csd_size )
SQATT   *sqatt ;
void    *csd ;
int     csd_size ;
{
     int     end_flag ;
     int     size_in_sock ;
     int     ioctl_rslt   ;
     int     read_rslt    ;
     int     read_len     ;
     int     rslt   ;
     int     widx   ;
     int     werrno ;
     int     old_sv_size ;
     int     first_flag  ;
     char    emsg[ 256 ] ;
     SQTBL    *sqtbl ;

     ttILog( &lhand , "[cs_recv()] Start. csd_size:%5d" , csd_size ) ;

     rslt  = 0 ;
     sqtbl = cur_sqtbl( sqatt ) ;
     old_sv_size    = sqtbl->sv_size ;
     sqtbl->ex_size = csd_size ;

     first_flag  = ( csd == 0x00 ) ? 0 : 1 ;
     /* About first_flag                                       */
     /* If first_flag equal to 1 ,  it is first try of reading */
     /* socket since notifying event.  If first_flag is 1  and */
     /* data size in socket is zero byte , it is regarded that */
     /* client socket has been already closed.                 */
     /* Then csd is 0x00  when reading body data from socket.  */
     /* And first_flag is always zero when reading body data.  */

     if( SQATT_BSIZE < csd_size )
     {
          sqtbl->ev_flag = 0 ;
          ttELog( &lhand ,
                  "Too long buffer required. SQATT:%5d , APP:%5d" ,
                  SQATT_BSIZE , csd_size ) ;

          return( SQATT_TOO_LONG_DATA_REQ ) ;
     }
     else if( csd_size <= sqtbl->sv_size )
     {
          /* There are data in SQATT internal buffer.          */
          /* But it is not received by user program.           */
          /* User program must receive it before next receive. */

          sqtbl->ev_flag = 0 ;
          ttELog( &lhand ,
                  "Buffer full. csd_size:%5d , sv_size:%5d" ,
                   csd_size , sqtbl->sv_size ) ;

          return( SQATT_BUFFER_FULL ) ;
     }

     end_flag = 0 ;

     while( end_flag == 0 )
     {
          size_in_sock = get_size_in_sock( sqatt ) ;

          if( size_in_sock == 0 )   /* No data in socket */
          {
               end_flag = 1 ;
               ttILog( &lhand , "No data in socket." ) ;
          }
          else if( size_in_sock > 0 )   /* Data exists in socket */
          {
               /* SQATT can read data of SQATT_BSIZE byte.                   */
               /* The value of csd_size is checked top of this function.     */
               /* It is garanteed that read_len is shorter than SQATT_BSIZE. */

               memset( sqtbl->tbuf , 0x00 , SQATT_BSIZE ) ;
               read_len  = csd_size - sqtbl->sv_size ;
               read_rslt = read( sqtbl->cfd , sqtbl->tbuf , read_len ) ;
               werrno    = errno ;

               if( read_rslt == -1 )
               {
                    ttELog( &lhand ,
                            "read() rslt:%5d , errno:%5d , %s" ,
                            read_rslt , werrno , strerror( werrno )) ;

                    rslt = SQATT_ERR ;
                    break ;
               }

               widx = sqtbl->sv_size ;
               memcpy( &sqtbl->mbuf[ widx ] , sqtbl->tbuf , read_rslt ) ;
               sqtbl->sv_size += read_rslt ;
               first_flag = 0 ;   /* Now , not first time.  */

               ttILog( &lhand , "data size(byte)" ) ;
               ttILog( &lhand , "expected:%5d , this time:%5d , total:%5d" ,
                                 csd_size , read_rslt , sqtbl->sv_size ) ;

               if( sqtbl->sv_size == csd_size )
               {
                    end_flag = 1 ;
               }
          }
     }

     if( rslt == 0 )
     {
          if( old_sv_size == sqtbl->sv_size )
          {
          /*   rslt = ( first_flag == 1 ) ? SQATT_CS_CLOSED : SQATT_NO_DATA ; */
               rslt = ( sqtbl->ev_flag == 1 )
                    ? SQATT_CS_CLOSED : SQATT_NO_DATA ;
          }
          else if( sqtbl->sv_size == csd_size )
          {
               if( csd != 0x00 )
               {
                    memcpy( csd , sqtbl->mbuf , csd_size ) ;
               }

               sqtbl->ex_size = 0 ;
               rslt = SQATT_OK ;
          }
          else if( 0 < sqtbl->sv_size && sqtbl->sv_size < csd_size )
          {
               rslt = SQATT_RECEIVING ;
          }
          else
          {
               rslt = SQATT_ERR ;
          }
     }

     ttILog( &lhand , "[cs_recv()] End. rslt:%4d" , rslt ) ;
     sqtbl->ev_flag = 0 ;

     return( rslt ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
int     sqatt_send( sqatt , sd_len , sd_data )
SQATT   *sqatt   ;
int     sd_len   ;
void    *sd_data ;
{
     int     rp_size ;
     int     rp_len  ;
     int     rslt ;

     ttILog( &lhand , "[sqatt_send()] Start. sd_len:%5d" , sd_len ) ;

     rp_size = 0 ;
     rp_len  = 0 ;

     rslt = sqatt_call( sqatt , sd_len , sd_data ,
                        rp_size , &rp_len , 0x00 ) ;

     ttILog( &lhand , "[sqatt_send()] End. rslt:%4d" , rslt ) ;

     return( rslt ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
int     sqatt_call( sqatt , sd_len , sd_data , rp_size , rp_len , rp_data )
SQATT   *sqatt   ;
int     sd_len   ;   /* rp_data has to be allocated memory            */
void    *sd_data ;   /* which is specified by rp_len.                 */
int     rp_size  ;   /* rp_data has to be allocated memory            */ 
int     *rp_len  ;   /* length of the data which was saved to rp_data */ 
void    *rp_data ;   /* which is specified by rp_len.                 */
{
     int     rslt    ;
     int     wr_rslt ;       /* result of write()        */
     int     rp_rslt ;       /* result of receive reply  */
     int     hd_len  ;       /* length of SQATT_HEAD     */
     int     werrno  ;

     char    w_dt[   20 ] ;  /* date & time in SQATT_HEAD */
     char    wbuf[ 1024 ] ;  /* for TEST */

     SQATT_HEAD    sqh ;
     SQTBL         *sqtbl ;

     ttILog( &lhand , "[sqatt_call()] Start." ) ;
     ttILog( &lhand , "- sd_len:%5d, rp_size:%5d, sq_type:%2d" ,
                       sd_len , rp_size , sqatt->sq_type ) ;

     if( sqatt->sq_type == 2 && rp_size == 0 )
     {
          /* In the case of 2 way queue, you must allocate space */
          /* to save reply message.  sqatt_call() returns error  */
          /* if there is no space for reply message.             */

          ttELog( &lhand ,
                  "No space has been allocated to save reply." ) ;
          ttELog( &lhand , "- sq_type: %6d" , sqatt->sq_type ) ;
          ttELog( &lhand , "- rp_len:  %6d" , rp_size ) ;

          rslt = SQATT_INVALID_PARAMETER ;

          return( rslt ) ;
     }

     /* edit sqatt header */
     edit_sqh( &sqh , sqatt , sd_len , 0 ) ;

     /* setup signal of SIGPIPE */
     signal( SIGPIPE , sqatt_signal ) ;

     /* set parameters of write() */
     hd_len  = sizeof( SQATT_HEAD ) ;
     sqtbl   = cur_sqtbl( sqatt ) ;

     /* send sqatt header */
     ttDILog( &lhand , &sqh , hd_len , "SQATT HEAD" ) ;
     wr_rslt = write( sqtbl->cfd , &sqh , hd_len ) ;
     werrno  = errno ;

     /* for TEST
     memset( wbuf , 0x00 , sizeof( wbuf )) ;
     memcpy( wbuf , &sqh , hd_len ) ;

     wr_rslt = write( sqtbl->cfd , &wbuf[ 0 ] , hd_len - 1 ) ;
     sleep( 10 ) ;
     wr_rslt = write( sqtbl->cfd , &wbuf[ hd_len - 1 ] , 1 ) ;
     wr_rslt = hd_len  ;
     werrno  = errno ;
     sleep( 10 ) ;
     /* for TEST */

     ttXLog( &lhand , "write() rslt:%5d, errno:%5d" , wr_rslt , werrno ) ;

     if( wr_rslt == hd_len )
     {
          ttILog( &lhand , "Completed send sqatt header." ) ;

          /* setup signal of SIGPIPE */
          signal( SIGPIPE , sqatt_signal ) ;

          /* send data part */
          ttDILog( &lhand , sd_data , sd_len , "SQATT BODY" ) ;
          if( sd_len > 0 )
          {
               wr_rslt = write( sqtbl->cfd , sd_data , sd_len ) ;
               werrno  = errno ;
          }
          else
          {
               wr_rslt = sd_len ;
               werrno  = 0 ;
          }

          /* for TEST
          memset( wbuf , 0x00    , sizeof( wbuf )) ;
          memcpy( wbuf , sd_data , sd_len ) ;

          wr_rslt = write( sqtbl->cfd , &wbuf[ 0 ] , sd_len - 1 ) ;
          sleep( 10 ) ;
          wr_rslt = write( sqtbl->cfd , &wbuf[ sd_len - 1 ] , 1 ) ;
          wr_rslt = sd_len ;
          werrno  = errno ;
          /* for TEST */

          ttXLog( &lhand , "write() rslt:%5d, errno:%5d" , wr_rslt , werrno ) ;

          if( wr_rslt == sd_len )
          {
               ttILog( &lhand , "Completed send body data." ) ;
               ttILog( &lhand , "recv_reply() ----------"   ) ;
               ttILog( &lhand , "rp_size:%5d" , rp_size     ) ;
                                                                                
               rp_rslt = recv_reply( sqatt , rp_size , rp_len , rp_data ) ;

               ttILog(  &lhand , "rp_len:%5d" , *rp_len ) ;
               ttDILog( &lhand , rp_data , *rp_len , "Reply Data" ) ;

               if( rp_rslt == SQATT_OK )
               {
                    countup_seqno( &sqtbl->seq_no ) ;
                    rslt = SQATT_OK ;

                    ttILog( &lhand , "Received reply. rslt:%4d" , rp_rslt ) ;
               }
               else
               {
                    ttELog( &lhand , "Error receiving reply. rslt:%4d" ,
                                      rp_rslt ) ;
                    rslt = rp_rslt ;
               }
          }
          else
          {
               ttELog( &lhand , "Data send Error. rslt:%4d , errno:%5d , %s" ,
                                 wr_rslt , werrno , strerror( werrno )) ;

               if( werrno = 32 ) /* Broken pipe */
               {
                    rslt = SQATT_BAD_QUEUE ;
               }
               else
               {
                    rslt = SQATT_ERR ;
               }
          }
     }
     else
     {
          ttELog( &lhand , "Header send error. rslt:%4d , errno:%5d , %s" ,
                            wr_rslt , werrno , strerror( werrno )) ;

          if( werrno == 32 ) /* Broken pipe */
          {
               rslt = SQATT_BAD_QUEUE ;
          }
          else
          {
               rslt = SQATT_ERR ;
          }
     }
                                                                                
     ttILog( &lhand , "[sqatt_call()] End." ) ;

     return( rslt ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  int     recv_reply( sqatt , sq_size , sq_len , sq_buf )
SQATT   *sqatt  ;
int     sq_size ;
int     *sq_len ;
void    *sq_buf ;
{
     int     end_flag ;
     int     fd_max   ;
     int     sqh_len  ;
     int     body_len ;
     int     select_rslt ;
     int     recv_rslt   ;
     int     rp_rslt     ;
     int     comp_rslt   ;
     int     rslt        ;

     fd_set  tfds , rfds ;
     SQTBL            *sqtbl ;
     SQATT_HEAD       sqh ;
     struct  timeval  *ptimer ;

     ttILog( &lhand , "[recv_reply] Start. sq_size:%5d" , sq_size ) ;

     sqtbl = cur_sqtbl( sqatt ) ;
     FD_ZERO( &tfds ) ;
     FD_ZERO( &rfds ) ;
     FD_SET ( sqtbl->cfd , &rfds ) ;

     *sq_len = 0 ;
     memset( sq_buf , 0x00 , sq_size ) ;

     memset( sqtbl->mbuf , 0x00 , SQATT_BSIZE ) ;
     sqatt->p_phase = SQATT_SELECT ;
     sqtbl->r_phase = RV_HEAD ;
     sqtbl->sv_size = 0 ;
     end_flag       = 0 ;

     while( end_flag == 0 )
     {
          if( sqatt->p_phase == SQATT_SELECT )
          {
               ttILog( &lhand , "[SQATT_SELECT] Start." ) ;

               fd_max = sqtbl->cfd + 1 ;
               ptimer = set_ctimeval( sqatt ) ;
               tfds   = rfds ;

               ttILog( &lhand, "Waiting reply ..." ) ;
               ttXLog( &lhand, "timeval = sec:%5d. usec:%5d , fd_max = %5d",
                       sqatt->ctimer.tv_sec , sqatt->ctimer.tv_usec ,
                       fd_max ) ;

               select_rslt = select( fd_max , &tfds ,
                                     ( fd_set* )0, ( fd_set* )0, ptimer ) ;

               if( select_rslt == 0 )
               {
                    ttELog( &lhand , "Timeout. No reply from server." ) ;
                    recv_rslt = SQATT_TIMEOUT ;
                    end_flag  = 1 ;
               }
               else if( 1 <= select_rslt )
               {
                    ttILog( &lhand , "Notified socket event. rslt:%4d" ,
                                     select_rslt ) ;

                    if( sqtbl->r_phase == RV_HEAD )
                    {
                         sqatt->p_phase = SQATT_RECV_HEAD ;
                    }
                    else
                    {
                         sqatt->p_phase = SQATT_RECV_BODY ;
                    }

                    end_flag = 0 ;
               }
               else
               {
                    ttELog( &lhand , "Error. recv_reply()/select" ) ;
                    recv_rslt = SQATT_ERR ;
                    end_flag  = 1 ;
               }

               ttILog( &lhand , "[SQATT_SELECT] End." ) ;
          }
          else if( sqatt->p_phase == SQATT_RECV_HEAD )
          {
               ttILog( &lhand , "[SQATT_RECV_HEAD] Start." ) ;

               sqh_len = sizeof( SQATT_HEAD ) ;
          /*   rp_rslt = rp_recv( sqatt , sqh_len , &sqh ) ; */
               rp_rslt = cs_recv( sqatt , &sqh , sqh_len ) ;
               ttILog( &lhand , "rp_rslt = %4d" , rp_rslt ) ;

               if( rp_rslt == SQATT_OK )
               {
                    ttDILog( &lhand , &sqh , sqh_len , "SQATT HEADER" ) ;

                    comp_rslt = check_sqhead( sqatt , &sqh ) ;

                    if( comp_rslt == SQATT_OK )
                    {
                         body_len = sqh.len - sqh_len ;
                         if( body_len == 0 )
                         {
                              if( sqatt->sq_type == 1 ) /* 1 way queue */
                              {
                                   recv_rslt = sqh.rslt ;
                                   end_flag  = 1 ;
                              }
                              else   /* 2 way queue */
                              {
                                   ttELog( &lhand , "No reply from server." ) ;
                                   recv_rslt = SQATT_NO_REPLY_SERVER ;
                                   end_flag  = 1 ;
                              }
                         }
                         else
                         {
                              if( sqatt->sq_type == 1 ) /* 1 way queue */
                              {
                                   ttELog( &lhand ,
                                           "Unexpected reply from server." ) ;
                                   recv_rslt = SQATT_UNEXPECTED_REPLY ;
                                   end_flag  = 1 ;
                              }
                              else   /* 2 way queue */
                              {
                                   ttILog( &lhand ,
                                           "Body part exsists. size:%4d bytes" ,
                                           body_len ) ;
                                   sqatt->p_phase = SQATT_SELECT ;
                                   sqtbl->r_phase = RV_BODY ;
                                   end_flag       = 0 ;
                              }
                         }
                    }
                    else
                    {
                         recv_rslt = comp_rslt ;
                         end_flag  = 1 ;
                    }

                    sqtbl_clrbuf( sqtbl ) ;
               }
               else if( rp_rslt == SQATT_RECEIVING )
               {
                    sqatt->p_phase = SQATT_SELECT ;
                    end_flag       = 0 ;
               }
               else
               {
                    sqtbl_clrbuf( sqtbl ) ;
                    recv_rslt = rp_rslt ;
                    end_flag  = 1 ;
               }

               ttILog( &lhand , "[SQATT_RECV_HEAD] End." ) ;
          }
          else if( sqatt->p_phase == SQATT_RECV_BODY )
          {
               ttILog( &lhand , "[SQATT_RECV_BODY] Start." ) ;

               if( sq_size >= body_len )
               {
               /*   rp_rslt = rp_recv( sqatt , body_len , sq_buf ) ; */
                    rp_rslt = cs_recv( sqatt , sq_buf , body_len ) ;
                    ttILog( &lhand , "rp_rslt = %4d" , rp_rslt ) ;

                    if( rp_rslt == SQATT_OK )
                    {
                         sqatt->p_phase = SQATT_SELECT ;
                         sqtbl->r_phase = RV_HEAD  ;
                         recv_rslt      = SQATT_OK ;
                         *sq_len        = body_len ;
                         end_flag       = 1 ;
                         sqtbl_clrbuf( sqtbl ) ;

                         ttXLog( &lhand , "body_len:%5d" , body_len ) ;
                    }
                    else if( rp_rslt == SQATT_RECEIVING )
                    {
                         sqatt->p_phase = SQATT_SELECT ;
                         end_flag       = 0 ;
                    }
                    else
                    {
                         sqtbl_clrbuf( sqtbl ) ;
                         recv_rslt = rp_rslt ;
                         end_flag  = 1 ;
                    }
               }
               else
               {
                    ttELog( &lhand , "Invalid reply data length." ) ;
                    ttELog( &lhand , "- expect:   %5d" , sq_len   ) ;
                    ttELog( &lhand , "- practice: %5d" , body_len ) ;

                    sqtbl_clrbuf( sqtbl ) ;
                    recv_rslt = SQATT_INVALID_LENGTH ;
                    end_flag  = 1 ;
               }

               ttILog( &lhand , "[SQATT_RECV_BODY] End." ) ;
          }
          else
          {
               ttELog( &lhand , "Invalid phase number. phase:%4d" ,
                                 sqatt->p_phase ) ;
               recv_rslt = SQATT_ERR ;
               end_flag  = 1 ;
          }
     }

     ttILog( &lhand , "[recv_reply] End." ) ;

     return( recv_rslt ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  int     rp_recv( sqatt , ilen , buf )
SQATT   *sqatt ;
int     ilen   ;
void    *buf   ;   /* This item has to be allocated enough memory. */
{
     int     recv_rslt ;
     int     read_rslt ;
     int     size_of_hope ;
     int     size_in_sock ;
     int     widx ;
     SQTBL   *sqtbl ;

     ttILog( &lhand , "[rp_recv] Start." ) ;

     recv_rslt = 0 ;
     sqtbl     = cur_sqtbl( sqatt ) ;

     size_of_hope = ilen - sqtbl->sv_size ;
     size_in_sock = get_size_in_sock( sqatt ) ;

     ttILog( &lhand , "rp_recv()" ) ;
     ttILog( &lhand , "- total      %4d bytes." , ilen ) ;
     ttILog( &lhand , "- this time  %4d bytes." , size_in_sock   ) ;
     ttILog( &lhand , "- accumulate %4d bytes." , sqtbl->sv_size ) ;

     if( size_in_sock == 0 )
     {
          /* Event notified but no data in socket. */
          /* Probabry socket disconnected.         */

          if( sqtbl->sv_size == 0 )
          {
               recv_rslt = SQATT_RECEIVING ;
          }
          else
          {
               recv_rslt = SQATT_CS_CLOSED ;
          }
     }
     else if( 0 < size_in_sock )
     {
          memset( sqtbl->tbuf , 0x00 , SQATT_BSIZE ) ;
          read_rslt = read( sqtbl->cfd , sqtbl->tbuf , size_in_sock ) ;
          ttILog( &lhand , "read() rslt:%4d" , read_rslt ) ;

          if( read_rslt == -1 )
          {
               recv_rslt = SQATT_ERR ;
          }
          else
          {
               widx = sqtbl->sv_size ;
               memcpy( &sqtbl->mbuf[ widx ] , sqtbl->tbuf , read_rslt ) ;
               sqtbl->sv_size += read_rslt ;

               if( sqtbl->sv_size == ilen )
               {
                    ttILog( &lhand , "Received data was saved in buf." ) ;
                    memset( buf , 0x00        , sqtbl->sv_size ) ;
                    memcpy( buf , sqtbl->mbuf , sqtbl->sv_size ) ;
               }
               else
               {
                    ttILog( &lhand , "Receiving data ..." ) ;
                    recv_rslt = SQATT_RECEIVING ;
               }
          }
     }
     else
     {
          ttILog( &lhand , "Too long or no data." ) ;
          recv_rslt = SQATT_TOO_LONG_REPLY ;
     }
                                                                                
     ttILog( &lhand , "[rp_recv] End." ) ;

     return( recv_rslt ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
int     sqatt_add_usrfd( sqatt , fd )
SQATT   *sqatt ;
int     fd ;
{
     int     widx  ;
     int     rslt  ;
     SQTBL    *sqtbl ;

     ttILog( &lhand , "[sqatt_add_usrfd()] Start." ) ;

     widx = get_blank_sqtbl( sqatt ) ;
     if( widx == -1 )
     {
          rslt = SQATT_FD_TBL_FULL ;
          ttELog( &lhand , "User fd table is full." ) ;
     }
     else
     {
          FD_SET( fd , &( sqatt->rfds )) ;
          sqtbl = &( sqatt->sqtbl[ widx ]) ;
          memset( sqtbl , 0x00 , sizeof( SQTBL )) ;

          sqtbl->fd_type = 1 ;  /* 1:user event       */
          sqtbl->cfd     = fd ; /* Save your event fd */
          rslt = 0 ;

          ttILog( &lhand , "User fd installed. fd:%4d" , sqtbl->cfd ) ;
     }

     ttILog( &lhand , "[sqatt_add_usrfd()] End." ) ;

     return( rslt ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
int     sqatt_del_usrfd( sqatt , fd )
SQATT   *sqatt ;
int     fd ;
{
     int     i ;
     int     find_idx  ;
     int     closed_fd ;
     int     rslt ;

     ttILog( &lhand , "[sqatt_del_usrfd()] Start." ) ;

     find_idx  = -1 ;
     closed_fd = fd ;

     for( i = 0 ; i < MAX_SQTBL ; ++i )
     {
          if( sqatt->sqtbl[ i ].cfd == fd )
          {
               find_idx = i ;
               break ;
          }
     }

     if( find_idx > -1 )
     {
          FD_CLR( fd , &( sqatt->rfds )) ;
          memset( &sqatt->sqtbl[ find_idx ] , 0x00 , sizeof( SQTBL )) ;

          ttILog( &lhand , "User fd deleted. fd:%3d" , closed_fd ) ;
          rslt = 0 ;
     }
     else
     {
          ttELog( &lhand , "User fd not found. fd:%3d" , closed_fd ) ;
          rslt = SQATT_ERR ;
     }

     ttILog( &lhand , "[sqatt_del_usrfd()] End." ) ;
     return( rslt ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
int     sqatt_get_usrfd( sqatt )
SQATT   *sqatt ;
{
     int     usr_fd ;
     SQTBL    *sqtbl ;

     ttILog( &lhand , "[sqatt_get_usrfd()] Start." ) ;
     sqtbl = cur_sqtbl( sqatt ) ;

     if( sqtbl->fd_type == 0 ) /* sqatt fd                             */
     {                         /* If this case happen , sqatt has bug. */
          ttELog( &lhand , "User fd doesn't exist." ) ;
          usr_fd = -1 ;
     }
     else   /* user fd */
     {
          ttILog( &lhand , "User fd found. fd:%4d" , sqtbl->cfd ) ;
          usr_fd = sqtbl->cfd ;
     } 

     ttILog( &lhand , "[sqatt_get_usrfd()] End." ) ;

     return( usr_fd ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  int     get_size_in_sock( sqatt )
SQATT   *sqatt ;
{
     int     ioctl_rslt   ;
     int     size_in_sock ;
     int     werrno ;
     SQTBL    *sqtbl ;

     sqtbl = cur_sqtbl( sqatt ) ;

     ioctl_rslt = ioctl( sqtbl->cfd , FIONREAD , &size_in_sock ) ;
     werrno = errno ;

     if( ioctl_rslt == -1 )
     {
          size_in_sock = 0 ;

          ttELog( &lhand ,
                  "ioctl() size:%5d , rslt:%5d , errno:%5d , %s" ,
                  size_in_sock , ioctl_rslt ,
                  werrno , strerror( werrno )) ;
     }

     return( size_in_sock ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
int     csh_send( sqatt , ret_code , body_len , sq_body )
SQATT   *sqatt ;
int     ret_code ;
int     body_len ;
void    *sq_body ;
{
     int     head_len ;
     int     werrno   ;
     int     fd_max   ;
     int     size_in_sock ;

     int     rslt ;
     int     wr_rslt  ;
     int     sel_rslt ;

     char    wbuf[ 4096 ] ;   /* TEST */

     fd_set        tfds    ;
     SQTBL         *sqtbl  ;
     SQATT_HEAD    sq_head ;

     struct  timeval  *timptr ;
     struct  timeval   timdat ;
                                                                                
     ttILog( &lhand , "[csh_send()] Start." ) ;
     ttXLog( &lhand , "ret_code:%5d , body_len:%5d , sq_body: 0x%08X" ,
                       ret_code , body_len , sq_body ) ;

     sqtbl = cur_sqtbl( sqatt ) ;

     /* select() */
     fd_max = sqtbl->cfd + 1 ;
     FD_ZERO( &tfds ) ;
     FD_SET( sqtbl->cfd , &tfds ) ;
     timdat.tv_sec  = 0 ;
     timdat.tv_usec = 0 ;
     timptr = &timdat ;

     sel_rslt = select( fd_max, &tfds, ( fd_set* )0, ( fd_set* )0, timptr ) ;
     werrno   = errno ;
     ttILog( &lhand , "select_rslt = %5d" , sel_rslt ) ;

     if( sel_rslt == 0 )
     {
          edit_sqh( &sq_head , sqatt , body_len , ret_code ) ;
          head_len = sizeof( SQATT_HEAD ) ;

          /* setup signal of SIGPIPE */
          signal( SIGPIPE , sqatt_signal ) ;

          /* log message */
          ttILog ( &lhand , "Sending header part ... " ) ;
          ttDILog( &lhand , &sq_head , head_len , "SQATT HEADER" ) ;

          /* Transmit header part */
          wr_rslt  = write( sqtbl->cfd , &sq_head , head_len ) ;
          werrno   = errno ;
          /* sent header */

          /* TEST
          memset( wbuf , 0x00     , sizeof( wbuf       )) ;
          memcpy( wbuf , &sq_head , sizeof( SQATT_HEAD )) ;

          wr_rslt = write( sqtbl->cfd , &wbuf[ 0 ] , 1 ) ;
          sleep( 10 ) ;
          wr_rslt = write( sqtbl->cfd , &wbuf[ 1 ] , head_len - 1 ) ;
          wr_rslt = head_len ;
          werrno  = errno ;
          sleep( 10 ) ;
          /* TEST     */

          ttXLog( &lhand , "write() rslt:%5d, errno:%5d" , wr_rslt , werrno ) ;

          if( wr_rslt == head_len )
          {
               ttILog( &lhand , "Completed sending header part." ) ;

               if( body_len == 0 )
               {
                    ttILog( &lhand , "And, no body part." ) ;

                    countup_seqno( &sqtbl->seq_no ) ;
                    rslt  = SQATT_OK ;
               }
               else
               {
                    /* log message */
                    ttILog ( &lhand , "Sending body part ... " ) ;
                    ttDILog( &lhand , sq_body , body_len , "SQATT BODY" ) ;

                    /* Transmit body part */
                    wr_rslt = write( sqtbl->cfd , sq_body , body_len ) ;
                    werrno  = errno ;
                    /* Transmitted body part */

                    /* TEST
                    memset( wbuf , 0x00    , sizeof( wbuf )) ;
                    memcpy( wbuf , sq_body , body_len ) ;

                    wr_rslt = write( sqtbl->cfd , &wbuf[ 0 ] , body_len - 1 ) ;
                    sleep( 10 ) ;
                    wr_rslt = write( sqtbl->cfd , &wbuf[ body_len - 1 ] , 1 ) ;
                    wr_rslt = body_len ;
                    werrno  = errno ;
                    /* TEST   */

                    ttXLog( &lhand ,
                            "write() rslt:%5d, errno:%5d" ,
                            wr_rslt , werrno ) ;

                    if( wr_rslt == body_len )
                    {
                         ttILog( &lhand , "Completed sending body part." ) ;

                         countup_seqno( &sqtbl->seq_no ) ;
                         rslt  = SQATT_OK ;
                    }
                    else
                    {
                         ttELog( &lhand ,
                                 "Error. rslt:%5d, errno:%5d , %s" ,
                                 wr_rslt , werrno , strerror( werrno )) ;

                         if( werrno == 32 )  /* Broken pipe */
                         {
                              rslt = SQATT_BAD_QUEUE ;
                         }
                         else
                         {
                              rslt = SQATT_ERR ;
                         }
                    }
               }
          }
          else
          {
               ttELog( &lhand ,
                       "Error. rslt:%5d, errno:%5d , %s" ,
                       wr_rslt , werrno , strerror( werrno )) ;

               if( werrno == 32 )  /* Broken pipe */
               {
                    rslt = SQATT_BAD_QUEUE ;
               }
               else
               {
                    rslt = SQATT_ERR ;
               }
          }
     }
     else if( sel_rslt > 0 )
     {
          if( FD_ISSET( sqtbl->cfd , &tfds ))
          {
               size_in_sock = get_size_in_sock( sqatt ) ;

               if( size_in_sock == 0 )
               {
                    /* Socket was closed by client process. */
                    ttELog( &lhand ,
                            "Client process closed socket." ) ;
                    rslt = SQATT_CS_CLOSED ;
               }
               else
               {
                    /* Data received before send reply. */
                    ttELog( &lhand , "Data received before reply." ) ;
                    rslt = SQATT_INVALID_SEQUENCE ;
               }
          }
          else
          {
               /* Unknown error. */
               ttELog( &lhand , "No match result select and FD_ISSET." ) ;
               rslt = SQATT_ERR ;
          }
     }
     else
     {
          /* select error */
          ttELog( &lhand , "Select error. rslt:%5d , errno:%5d , %s" ,
                   sel_rslt , werrno , strerror( werrno )) ;
          rslt = SQATT_ERR ;
     }

     ttILog( &lhand , "[csh_send()] End. rslt:%5d" , rslt ) ;

     return( rslt ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  void    edit_sqh( sqh , sqatt , body_len , result_code )
SQATT_HEAD      *sqh        ;
SQATT           *sqatt      ;
int             body_len    ;
int             result_code ;
{
     int        head_len  ;
     char       wdt[ 20 ] ;
     SQTBL      *sqtbl    ;

     sqtbl    = cur_sqtbl( sqatt ) ;
     head_len = sizeof( SQATT_HEAD ) ;
     memset( sqh , 0x20 , head_len ) ;

     memset ( wdt ,  0x00  , sizeof( wdt )) ;
     sprintf( wdt , "%s%s" , ttDate4( 0x00 ) , ttTime( 0x00 )) ;

     sqh->len     = head_len + body_len ;
     sqh->version = SQ_VERSION     ;
     sqh->sq_type = sqatt->sq_type ;
     sqh->seq_no  = sqtbl->seq_no  ;
     sqh->rslt    = result_code    ;

     memset( sqh->msg_cd , '0' , sizeof( sqh->msg_cd )) ;
     memcpy( sqh->lhdt   , wdt , sizeof( sqh->lhdt   )) ;
}

/* ********************************************************************** */
/* * set_fd_max()                                                       * */
/* *                                                                    * */
/* * Search file descriptors having maximum value, and save it.         * */
/* *                                                                    * */
/* ********************************************************************** */
static  int     set_fd_max( sqatt )
SQATT   *sqatt ;
{
     int     i ;
     int     fd_max ;

     fd_max = sqatt->sfd ;

     for( i = 0 ; i < MAX_SQTBL ; ++i )
     {
          if( fd_max < sqatt->sqtbl[ i ].cfd )
          {
               fd_max = sqatt->sqtbl[ i ].cfd ;
          }
     }

     fd_max += 1 ;

     return( fd_max ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  int     check_sqhead( sqatt , sqh )
SQATT           *sqatt ;
SQATT_HEAD      *sqh ;
{
     int     recv_no ;
     int     save_no ;
     int     rslt    ;
     SQTBL   *sqtbl  ;

     rslt  = SQATT_OK ;
     sqtbl = cur_sqtbl( sqatt ) ;

     do
     {
          if( sqh->version == SQ_VERSION )
          {
               ttILog( &lhand , "SQATT version is %02d." , sqh->version ) ;
               rslt = SQATT_OK ;
          }
          else
          {
               ttELog( &lhand ,
                       "Invalid SQATT version. server:%02d , client:%02d" ,
                        SQ_VERSION , sqh->version ) ;
               rslt = SQATT_INVALID_VERSION ;
               break ;
          }

          if( sqatt->sq_type == sqh->sq_type )
          {
               ttILog( &lhand , "sq_type is %1d." , sqh->sq_type ) ;
               rslt = SQATT_OK ;
          }
          else
          {
               ttELog( &lhand ,
                       "Conflicted sq_type. server:%02d , client:%02d" ,
                       sqatt->sq_type , sqh->sq_type ) ;
               rslt = SQATT_CONFLICTED_SQ_TYPE ;
               break ;
          }

          recv_no = sqh->seq_no ;
          save_no = sqtbl->seq_no ;

          if( recv_no == save_no )
          {
               ttILog( &lhand , "Current seq_no is %012d." , recv_no ) ;
               rslt = SQATT_OK ;
          }
          else if( recv_no > save_no )
          {
               ttELog( &lhand ,
                       "Data lost. Expected:%012d , Received:%012d" ,
                        save_no , recv_no ) ;
               rslt = SQATT_SKIP_SEQ_NO ;
               break ;
          }
          else
          {
               ttELog( &lhand ,
                       "Data duplivated. Expected:%012d , Received:%012d" ,
                        save_no , recv_no ) ;
               rslt = SQATT_DUP_SEQ_NO ;
               break ;
          }
     }
     while( 0 ) ;

     return( rslt ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  struct  timeval  *set_stimeval( sqatt )
SQATT   *sqatt ;
{
     struct  timeval  *ptimer ;

     if( sqatt->stimer.tv_sec  == 0 &&
         sqatt->stimer.tv_usec == 0 )
     {
          /* In this case, select() waits for event forever. */
          ptimer = 0x00 ;
     }
     else
     {
          ptimer = &( sqatt->stimer ) ;
     }

     return( ptimer ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  struct  timeval  *set_ctimeval( sqatt )
SQATT   *sqatt ;
{
     struct  timeval  *ptimer ;

     if( sqatt->ctimer.tv_sec  == 0 &&
         sqatt->ctimer.tv_usec == 0 )
     {
          /* In this case, select() waits for event forever. */
          ptimer = 0 ;
     }
     else
     {
          ptimer = &( sqatt->ctimer ) ;
     }

     return( ptimer ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  int     sqtbl_clrbuf( sqtbl )
SQTBL   *sqtbl ;
{
     memset( sqtbl->mbuf , 0x00 , SQATT_BSIZE ) ;
     memset( sqtbl->tbuf , 0x00 , SQATT_BSIZE ) ;
     sqtbl->sv_size = 0 ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  int     clear_sqtbl( sqtbl )
SQTBL   *sqtbl ;
{
     memset( sqtbl , 0x00 , sizeof( SQTBL )) ;

     sqtbl->fd_type =  0 ;  /* User fd                     */
     sqtbl->ev_flag =  0 ;  /* Not received event          */
     sqtbl->cfd     = -1 ;  /* Initial value of cfd.       */
     sqtbl->r_phase = RV_HEAD ;
     sqtbl->sv_size =  0 ;  /* Saved data size             */
     sqtbl->bd_size =  0 ;  /* Body size                   */
     sqtbl->ex_size =  0 ;  /* expept size of receive data */
     sqtbl->seq_no  =  0 ;  /* sequential number           */
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  int     get_blank_sqtbl( sqatt )
SQATT   *sqatt ;
{
     int     i ;
     int     sqtbl_idx ;

     sqtbl_idx = -1 ;
     dump_cfd( sqatt ) ;

     for( i = 0 ; i < MAX_SQTBL ; ++i )
     {
          if( sqatt->sqtbl[ i ].cfd == -1 )
          {
               sqtbl_idx = i ;
               break ;
          }
     }

     return( sqtbl_idx ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  SQTBL   *cur_sqtbl( sqatt )
SQATT   *sqatt ;
{
     return( &( sqatt->sqtbl[ sqatt->cur_idx ])) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  void    countup_seqno( cnt )
int     *cnt ;
{
     int       i ;
     int       mcnt ;
     int       old_cnt ;
     unsigned  int     max ;

     old_cnt = *cnt ;
     max     = 1 ;
     mcnt    = sizeof( int ) * 8  - 1 ;

     for( i = 0 ; i < mcnt ; ++i )
     {
          max = max * 2 ;
     }

     if( *cnt < max - 1 )
     {
          *cnt += 1 ;
     }
     else
     {
          *cnt = 0 ;
     }

     ttILog( &lhand , "countup_seqno()  %08d -> %08d" ,
                       old_cnt , *cnt ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  void    dump_sqh( sqh )
SQATT_HEAD      *sqh ;
{
     int     sqh_size ;
     int     sqb_size ;

     char    w_cd  [ 10 ] ;
     char    w_lhdt[ 20 ] ;

     sqh_size = sizeof( SQATT_HEAD ) ;
     sqb_size = sqh->len - sqh_size  ;

     memset( w_lhdt , 0x00 , sizeof( w_lhdt )) ;
     memset( w_cd   , 0x00 , sizeof( w_cd   )) ;

     memcpy( w_cd   , sqh->msg_cd , sizeof( sqh->msg_cd )) ;
     memcpy( w_lhdt , sqh->lhdt   , sizeof( sqh->lhdt   )) ;

     ttILog ( &lhand , "[SQATT HEADER INFORMATION]" ) ;
     ttILog ( &lhand , "header:    %6d"  , sqh_size       ) ;
     ttILog ( &lhand , "body:      %6d"  , sqb_size       ) ;
     ttILog ( &lhand , "version:   %03d" , sqh->version   ) ;
     ttILog ( &lhand , "seq_no:    %06d" , sqh->seq_no    ) ;
     ttILog ( &lhand , "date time: %s"   , w_lhdt         ) ;
     ttILog ( &lhand , "msg_cd:    %s"   , w_cd           ) ;
     ttDILog( &lhand , sqh , sqh_size    , "SQATT HEADER" ) ;
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
static  void    dump_cfd( sqatt )
SQATT   *sqatt ;
{
     int     i ;
     int     cnt ;

     cnt = 0 ;
/*   ttXLog( &lhand , "[dump_cfd()] Start. sqatt: 0x%08X" , sqatt ) ; */

     for( i = 0 ; i < MAX_SQTBL ; ++i )
     {
          if( sqatt->sqtbl[ i ].cfd > 0 )
          {
          /*   ttXLog( &lhand , "cfd[%4d] = %4d" ,             */
          /*                     i , sqatt->sqtbl[ i ].cfd ) ; */
               cnt += 1 ;
          }
     }

/*   ttXLog( &lhand , "Connected client process is %4d." , cnt ) ; */
/*   ttXLog( &lhand , "[dump_cfd()] End." ) ;                      */
}

/* ********************************************************************** */
/* *                                                                    * */
/* ********************************************************************** */
void    sqatt_signal( int sig )
{
     ttELog( &lhand , "[sqatt_signal()] Signal caught. SigNo:%5d" , sig ) ;
     signal( SIGPIPE , SIG_DFL ) ;
}
